import numpy as np
from lime.lime_image import LimeImageExplainer
from lime.lime_text import LimeTextExplainer
from dianna import utils


class LIME:
    """
    LIME implementation as wrapper around https://github.com/marcotcr/lime
    """
    # axis labels required to be present in input image data
    required_labels = ('batch', 'channels')

    def __init__(self,
                 kernel_width=25,
                 kernel=None,
                 verbose=False,
                 class_names=None,
                 feature_selection='auto',
                 split_expression=r'\W+',
                 bow=False,
                 mask_string=None,
                 random_state=None,
                 char_level=False,
                 axes_labels=None,
                 preprocess_function=None,
                 ):  # pylint: disable=too-many-arguments
        """LIME initializer.

        Args:
            kernel_width (int, optional):
            kernel (callable, optional):
            verbose (bool, optional):
            class_names (list, optional):
            feature_selection (str, optional):
            split_expression (regexp, optional):
            bow (bool, optional):
            mask_string (str, optional):
            random_state (int or np.RandomState, optional):
            char_level (bool, optional):
            axes_labels (dict/list, optional): If a dict, key,value pairs of axis index, name.
                                               If a list, the name of each axis where the index
                                               in the list is the axis index
            preprocess_function (callable, optional): Function to preprocess input data with
        """
        self.text_explainer = LimeTextExplainer(kernel_width,
                                                kernel,
                                                verbose,
                                                class_names,
                                                feature_selection,
                                                split_expression,
                                                bow,
                                                mask_string,
                                                random_state,
                                                char_level,
                                                )

        self.image_explainer = LimeImageExplainer(kernel_width,
                                                  kernel,
                                                  verbose,
                                                  feature_selection,
                                                  random_state,
                                                  )

        self.preprocess_function = preprocess_function
        self.axes_labels = axes_labels if axes_labels is not None else []

    def explain_text(self,
                     model_or_function,
                     input_data,
                     labels=(0,),
                     top_labels=None,
                     num_features=10,
                     num_samples=5000,
                     **kwargs,
                     ):  # pylint: disable=too-many-arguments
        """
        Run the LIME explainer.

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_data (np.ndarray): Data to be explained
            labels ([int], optional): Iterable of indices of class to be explained

        Other keyword arguments: see the LIME documentation for LimeTextExplainer.explain_instance:
        https://lime-ml.readthedocs.io/en/latest/lime.html#lime.lime_text.LimeTextExplainer.explain_instance.

        Returns:
            list of (word, index of word in raw text, importance for target class) tuples
        """
        runner = utils.get_function(model_or_function, preprocess_function=self.preprocess_function)
        explain_instance_kwargs = utils.get_kwargs_applicable_to_function(self.text_explainer.explain_instance, kwargs)
        explanation = self.text_explainer.explain_instance(input_data,
                                                           runner,
                                                           labels=labels,
                                                           top_labels=top_labels,
                                                           num_features=num_features,
                                                           num_samples=num_samples,
                                                           **explain_instance_kwargs
                                                           )

        local_explanations = explanation.local_exp
        string_map = explanation.domain_mapper.indexed_string
        return [self._get_results_for_single_label(local_explanations[label], string_map) for label in labels]

    @staticmethod
    def _get_results_for_single_label(local_explanation, string_map):
        return [(string_map.word(index), int(string_map.string_position(index)), importance)
                for index, importance in local_explanation]

    def explain_image(self,
                      model_or_function,
                      input_data,
                      label=1,
                      top_labels=None,
                      num_features=10,
                      num_samples=5000,
                      positive_only=False,
                      hide_rest=True,
                      **kwargs,
                      ):  # pylint: disable=too-many-arguments,too-many-locals
        """
        Run the LIME explainer.

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_data (np.ndarray): Data to be explained
            label (int): Index of class to be explained
        Other keyword arguments: see the LIME documentation for LimeImageExplainer.explain_instance and
        ImageExplanation.get_image_and_mask:

        - https://lime-ml.readthedocs.io/en/latest/lime.html#lime.lime_image.LimeImageExplainer.explain_instance
        - https://lime-ml.readthedocs.io/en/latest/lime.html#lime.lime_image.ImageExplanation.get_image_and_mask

        Returns:
            list of (word, index of word in raw text, importance for target class) tuples
        """
        input_data, full_preprocess_function = self._prepare_image_data(input_data)
        runner = utils.get_function(model_or_function, preprocess_function=full_preprocess_function)

        # run the explanation.
        explain_instance_kwargs = utils.get_kwargs_applicable_to_function(self.image_explainer.explain_instance, kwargs)
        explanation = self.image_explainer.explain_instance(input_data,
                                                            runner,
                                                            labels=(label,),
                                                            top_labels=top_labels,
                                                            num_features=num_features,
                                                            num_samples=num_samples,
                                                            **explain_instance_kwargs,
                                                            )

        get_image_and_mask_kwargs = utils.get_kwargs_applicable_to_function(explanation.get_image_and_mask, kwargs)
        mask = explanation.get_image_and_mask(label, positive_only=positive_only, hide_rest=hide_rest,
                                              num_features=num_features, **get_image_and_mask_kwargs)[1]
        return mask

    def _prepare_image_data(self, input_data):
        """
        Transforms the data to be of the shape and type LIME expects

        Args:
            input_data (NumPy-compatible array): Data to be explained

        Returns:
            transformed input data, preprocessing function to use with utils.get_function()
        """
        input_data = utils.to_xarray(input_data, self.axes_labels, LIME.required_labels)
        # remove batch axis from input data; this is only here for a consistent API
        # but LIME wants data without batch axis
        if not len(input_data['batch']) == 1:
            raise ValueError("Length of batch axis must be one.")
        input_data = input_data.sel(batch=0)
        # ensure channels axis is last and keep track of where it was so we can move it back
        channels_axis_index = input_data.dims.index('channels')
        input_data = utils.move_axis(input_data, 'channels', -1)
        # create preprocessing function that puts model input generated by LIME into the right shape and dtype,
        # followed by running the user's preprocessing function
        full_preprocess_function = self._get_full_preprocess_function(channels_axis_index, input_data.dtype)
        # LIME requires float64 numpy data
        return input_data.values.astype(np.float64), full_preprocess_function

    def _get_full_preprocess_function(self, channel_axis_index, dtype):
        """
        Create a preprocessing function that incorporates both the (optional) user's
        preprocessing function, as well as any needed dtype and shape conversions

        Args:
            channel_axis_index (int): Axis index of the channels in the input data
            dtype (type): Data type of the input data (e.g. np.float32)

        Returns:
            Function that first ensures the data has the same shape and type as the input data,
            then runs the users' preprocessing function
        """
        # LIME generates numpy arrays, so it is easiest here to use np.transpose to put the channels
        # axis in the right place
        # one is added to the channels axis index because there is an extra first axis: the batch axis
        def moveaxis_function(data):
            return np.moveaxis(data, -1, channel_axis_index + 1).astype(dtype)

        if self.preprocess_function is None:
            return moveaxis_function
        return lambda data: self.preprocess_function(moveaxis_function(data))



import numpy as np
from skimage.transform import resize
from tqdm import tqdm
from dianna import utils


def normalize(saliency, n_masks, p_keep):
    return saliency / n_masks / p_keep


def _upscale(grid_i, up_size):
    return resize(grid_i, up_size, order=1, mode='reflect', anti_aliasing=False)


class RISE:
    """
    RISE implementation based on https://github.com/eclique/RISE/blob/master/Easy_start.ipynb
    """
    # axis labels required to be present in input image data
    required_labels = ('batch', 'channels')

    def __init__(self, n_masks=1000, feature_res=8, p_keep=0.5,  # pylint: disable=too-many-arguments
                 axes_labels=None, preprocess_function=None):

        """RISE initializer.

        Args:
            n_masks (int): Number of masks to generate.
            feature_res (int): Resolution of features in masks.
            p_keep (float): Fraction of image to keep in each mask
            axes_labels (dict/list, optional): If a dict, key,value pairs of axis index, name.
                                               If a list, the name of each axis where the index
                                               in the list is the axis index
            preprocess_function (callable, optional): Function to preprocess input data with
        """
        self.n_masks = n_masks
        self.feature_res = feature_res
        self.p_keep = p_keep
        self.preprocess_function = preprocess_function
        self.masks = None
        self.predictions = None
        self.axes_labels = axes_labels if axes_labels is not None else []

    def explain_text(self, model_or_function, input_text, labels=(0,), batch_size=100):
        runner = utils.get_function(model_or_function, preprocess_function=self.preprocess_function)
        input_tokens = np.asarray(model_or_function.tokenizer(input_text))
        text_length = len(input_tokens)
        p_keep = self._determine_p_keep_for_text(input_tokens, runner) if self.p_keep is None else self.p_keep
        input_shape = (text_length,)
        self.masks = self._generate_masks_for_text(input_shape, p_keep,
                                                   self.n_masks)  # Expose masks for to make user inspection possible
        sentences = self._create_masked_sentences(input_tokens, self.masks)
        saliencies = self._get_saliencies(runner, sentences, text_length, batch_size, p_keep)
        return self._reshape_result(input_tokens, labels, saliencies)

    def _determine_p_keep_for_text(self, input_data, runner, n_masks=100):
        """
        See n_mask default value https://github.com/dianna-ai/dianna/issues/24#issuecomment-1000152233
        """
        p_keeps = np.arange(0.1, 1.0, 0.1)
        stds = []
        for p_keep in p_keeps:
            std = self._calculate_mean_class_std_for_text(p_keep, runner, input_data, n_masks=n_masks)
            stds += [std]
        best_i = np.argmax(stds)
        best_p_keep = p_keeps[best_i]
        print(f'Rise parameter p_keep was automatically determined at {best_p_keep}')
        return best_p_keep

    def _calculate_mean_class_std_for_text(self, p_keep, runner, input_data, n_masks):
        batch_size = 50
        masks = self._generate_masks_for_text(input_data.shape, p_keep, n_masks)
        masked = self._create_masked_sentences(input_data, masks)
        predictions = []
        for i in range(0, n_masks, batch_size):
            current_input = masked[i:i + batch_size]
            current_predictions = runner(current_input)
            predictions.append(current_predictions)
        predictions = np.concatenate(predictions)
        std_per_class = predictions.std(axis=0)
        return np.mean(std_per_class)

    def _generate_masks_for_text(self, input_shape, p_keep, n_masks):
        masks = np.random.choice(a=(True, False), size=(n_masks,) + input_shape, p=(p_keep, 1 - p_keep))
        return masks

    def _get_saliencies(self, runner, sentences, text_length, batch_size, p_keep):  # pylint: disable=too-many-arguments
        self.predictions = self._get_predictions(sentences, runner, batch_size)
        unnormalized_saliency = self.predictions.T.dot(self.masks.reshape(self.n_masks, -1)).reshape(-1, text_length)
        return normalize(unnormalized_saliency, self.n_masks, p_keep)

    @staticmethod
    def _reshape_result(input_tokens, labels, saliencies):
        word_lengths = [len(t) for t in input_tokens]
        word_indices = [sum(word_lengths[:i]) + i for i in range(len(input_tokens))]
        return [list(zip(input_tokens, word_indices, saliencies[label])) for label in labels]

    def _get_predictions(self, sentences, runner, batch_size):
        predictions = []
        for i in tqdm(range(0, self.n_masks, batch_size), desc='Explaining'):
            predictions.append(runner(sentences[i:i + batch_size]))
        predictions = np.concatenate(predictions)
        return predictions

    def _create_masked_sentences(self, tokens, masks):
        tokens_masked = []
        for mask in masks:
            tokens_masked.append(tokens[mask])
        sentences = [" ".join(t) for t in tokens_masked]
        return sentences

    def explain_image(self, model_or_function, input_data, batch_size=100):
        """Run the RISE explainer.
           The model will be called with masked images,
           with a shape defined by `batch_size` and the shape of `input_data`

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_data (np.ndarray): Image to be explained
            batch_size (int): Batch size to use for running the model.

        Returns:
            Explanation heatmap for each class (np.ndarray).
        """

        runner = utils.get_function(model_or_function, preprocess_function=self.preprocess_function)
        # convert data to xarray
        input_data = utils.to_xarray(input_data, self.axes_labels, RISE.required_labels)
        # batch axis should always be first
        input_data = utils.move_axis(input_data, 'batch', 0)
        # ensure channels axis is last and keep track of where it was so we can move it back
        channels_axis_index = input_data.dims.index('channels')
        input_data = utils.move_axis(input_data, 'channels', -1)

        p_keep = self._determine_p_keep_for_images(input_data, runner) if self.p_keep is None else self.p_keep

        # data shape without batch axis and channel axis
        img_shape = input_data.shape[1:3]
        # Expose masks for to make user inspection possible
        self.masks = self.generate_masks_for_images(img_shape, p_keep, self.n_masks)

        # Make sure multiplication is being done for correct axes
        masked = (input_data * self.masks)
        # ensure channels axis is in original location again
        masked = utils.move_axis(masked, 'channels', channels_axis_index)
        # convert to numpy for onnx
        masked = masked.values.astype(input_data.dtype)

        batch_predictions = []
        for i in tqdm(range(0, self.n_masks, batch_size), desc='Explaining'):
            batch_predictions.append(runner(masked[i:i + batch_size]))
        self.predictions = np.concatenate(batch_predictions)

        saliency = self.predictions.T.dot(self.masks.reshape(self.n_masks, -1)).reshape(-1, *img_shape)
        return normalize(saliency, self.n_masks, p_keep)

    def _determine_p_keep_for_images(self, input_data, runner, n_masks=100):
        """
        See n_mask default value https://github.com/dianna-ai/dianna/issues/24#issuecomment-1000152233
        """
        p_keeps = np.arange(0.1, 1.0, 0.1)
        stds = []
        for p_keep in p_keeps:
            std = self._calculate_mean_class_std_for_images(p_keep, runner, input_data, n_masks=n_masks)
            stds += [std]
        best_i = np.argmax(stds)
        best_p_keep = p_keeps[best_i]
        print(f'Rise parameter p_keep was automatically determined at {best_p_keep}')
        return best_p_keep

    def _calculate_mean_class_std_for_images(self, p_keep, runner, input_data, n_masks):
        batch_size = 50
        img_shape = input_data.shape[1:3]
        masks = self.generate_masks_for_images(img_shape, p_keep, n_masks)
        masked = (input_data * masks).astype(input_data.dtype)
        predictions = []
        for i in range(0, n_masks, batch_size):
            current_input = masked[i:i + batch_size]
            current_predictions = runner(current_input)
            predictions.append(current_predictions)
        predictions = np.concatenate(predictions)
        std_per_class = predictions.std(axis=0)
        return np.mean(std_per_class)

    def generate_masks_for_images(self, input_size, p_keep, n_masks):
        """Generate a set of random masks to mask the input data

        Args:
            input_size (int): Size of a single sample of input data, for images without the channel axis.
        Returns:
            The generated masks (np.ndarray)
        """
        cell_size = np.ceil(np.array(input_size) / self.feature_res)
        up_size = (self.feature_res + 1) * cell_size

        grid = np.random.choice(a=(True, False), size=(n_masks, self.feature_res, self.feature_res),
                                p=(p_keep, 1 - p_keep))
        grid = grid.astype('float32')

        masks = np.empty((n_masks, *input_size))

        for i in range(n_masks):
            y = np.random.randint(0, cell_size[0])
            x = np.random.randint(0, cell_size[1])
            # Linear upsampling and cropping
            masks[i, :, :] = _upscale(grid[i], up_size)[y:y + input_size[0], x:x + input_size[1]]
        masks = masks.reshape(-1, *input_size, 1)
        return masks



"""Documentation about dianna"""
import logging
from . import methods


logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "DIANNA Team"
__email__ = "dianna-ai@esciencecenter.nl"
__version__ = "0.2.0"


def explain_image(model_or_function, input_data, method, **kwargs):
    """
    Exampler explainer wrapper
    """
    return get_explainer(method, kwargs).explain_image(model_or_function, input_data)


def explain_text(model_or_function, input_data, method, labels=(1,), **kwargs):
    """
    Exampler explainer wrapper
    """
    return get_explainer(method, kwargs).explain_text(model_or_function, input_data, labels)


def get_explainer(method, kwargs):
    method_class = getattr(methods, method)
    return method_class(**kwargs)



import numpy as np


class SHAP:
    """
    SHAP implementation based on captum https://github.com/pytorch/captum
    """
    def __init__(self, shap_arg_1=0, shap_arg_2=0):
        """
        Example constructor for testing.
        """
        self.shap_arg_1 = shap_arg_1
        self.shap_arg_2 = shap_arg_2

    def explain_image(self, model, input_data):
        """
        Example call function.
        """
        self.model = model

        return np.zeros(input_data.shape, dtype=float)




import xarray as xr
from dianna.utils.onnx_runner import SimpleModelRunner


def get_function(model_or_function, preprocess_function=None):
    """Converts input to callable function.
       Input can be either model path or function.
       If input is a function, the function is returned unchanged.

       Any keyword arguments are given to the ModelRunner class if the input is a model path.
    """
    if isinstance(model_or_function, str):
        runner = SimpleModelRunner(model_or_function, preprocess_function=preprocess_function)
    elif callable(model_or_function):
        if preprocess_function is None:
            runner = model_or_function
        else:
            def runner(input_data):
                return model_or_function(preprocess_function(input_data))
    else:
        raise TypeError("model_or_function argument must be string (path to model) or function")
    return runner


def get_kwargs_applicable_to_function(function, kwargs):
    """
    Returns a dict that is the subset of `kwargs` for which the keys are
    keyword arguments of `function`. Note that if `function` has a `**kwargs`
    argument, this function should not be necessary (provided the function
    handles `**kwargs` robustly).
    """
    return {key: value for key, value in kwargs.items()
            if key in function.__code__.co_varnames}


def to_xarray(data, axes_labels, required_labels=None):
    """Converts numpy data and axes labels to an xarray object
    """
    if isinstance(axes_labels, dict):
        # key = axis index, value = label
        # not all axes have to be present in the input, but we need to provide
        # a name for each axis
        # first ensure negative indices are converted to positive ones
        indices = list(axes_labels.keys())
        for index in indices:
            if index < 0:
                axes_labels[data.ndim + index] = axes_labels.pop(index)
        labels = [axes_labels[index] if index in axes_labels else f'dim_{index}' for index in range(data.ndim)]
    else:
        labels = list(axes_labels)

    # check if the required labels are present
    if required_labels is not None:
        for label in required_labels:
            assert label in labels, f'Required label missing: {label}'

    return xr.DataArray(data, dims=labels)


def move_axis(data, label, new_position):
    """Moves a named axis to a new position in an xarray DataArray object.

    Args:
        data (DataArray): Object with named axes
        label (str): Name of the axis to move
        new_position (int): Numerical new position of the axis.
                            Negative indices are accepted.

    Returns:
        data with axis in new position
    """
    # find current position of axis
    try:
        pos = data.dims.index(label)
    except ValueError as e:
        raise ValueError(f"Axis name {label} does not exist in input data") from e

    # create list of labels with new ordering
    axis_labels = list(data.dims)
    # the new position will be _before_ the given index, so will fail with a negative index
    # convert to a positive index in that case
    if new_position < 0:
        new_position += len(axis_labels)
    axis_labels.insert(new_position, axis_labels.pop(pos))
    # do the move
    return data.transpose(*axis_labels)
